use std::sync::Arc;

use askama::Template;
use axum::{
    debug_handler,
    extract::State,
    response::IntoResponse,
    routing::{get, post},
    Form, Router,
};
use serde::Deserialize;
use tokio::{net::TcpListener, sync::Mutex};
use tower_http::services::ServeDir;
use tracing::info;
use tracing_subscriber::{layer::SubscriberExt, util::SubscriberInitExt};

#[warn(clippy::all)]
#[warn(clippy::pedantic)]
#[tokio::main]
async fn main() {
    tracing_subscriber::registry()
        .with(
            tracing_subscriber::EnvFilter::try_from_default_env()
                .unwrap_or_else(|_| "fastmx=info".into()),
        )
        .with(tracing_subscriber::fmt::layer())
        .init();

    let pwd = std::env::current_dir().unwrap();

    let todo_state = Arc::new(TodoState {
        todos: Mutex::new(Vec::new()),
    });

    let api_router = Router::new()
        .route("/hello", get(hello_from_the_server))
        .route("/goodbye", get(goodbye))
        .route("/todos", post(add_todo))
        .with_state(todo_state);

    let router = Router::new()
        .nest("/api", api_router)
        .route("/", get(hello))
        .route("/another-page", get(another_page))
        .nest_service(
            "/assets",
            ServeDir::new(format!("{}/assets", pwd.to_str().unwrap())),
        );

    let listener = get_listener().await;
    info!("listening on http://{}", listener.local_addr().unwrap());
    axum::serve(listener, router).await.unwrap();
}

async fn get_listener() -> TcpListener {
    let mut port = 3000;
    loop {
        match TcpListener::bind(format!("0.0.0.0:{port}")).await {
            Ok(listener) => return listener,
            Err(_) => port += 1,
        }
    }
}

async fn hello() -> impl IntoResponse {
    HelloTemplate {}
}

async fn another_page() -> impl IntoResponse {
    AnotherPageTemplate {}
}

async fn hello_from_the_server() -> impl IntoResponse {
    "Hello from the server"
}

async fn goodbye() -> impl IntoResponse {
    "goodbye"
}

#[debug_handler]
async fn add_todo(
    State(todo_state): State<Arc<TodoState>>,
    Form(todo): Form<TodoRequest>,
) -> impl IntoResponse {
    todo_state.todos.lock().await.push(todo.name);
    TodoListTemplate {
        todos: todo_state.todos.lock().await.clone(),
    }
}

#[derive(Deserialize, Debug)]
struct TodoRequest {
    name: String,
}

#[derive(Template)]
#[template(path = "hello.html")]
struct HelloTemplate {}

#[derive(Template)]
#[template(path = "another-page.html")]
struct AnotherPageTemplate {}

#[derive(Template)]
#[template(path = "todo-list.html")]
struct TodoListTemplate {
    todos: Vec<String>,
}

struct TodoState {
    todos: Mutex<Vec<String>>,
}
